/**
 * Copyright (C) 2010 Google Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.inject.assistedinject;

import com.google.inject.spi.BindingTargetVisitor;

/**
 * A visitor for the AssistedInject extension.
 * <p>
 * If your {@link BindingTargetVisitor} implements this interface, bindings created by using
 * {@link FactoryModuleBuilder} will be visited through this interface.
 *
 * @since 3.0
 * @author ramakrishna@google.com (Ramakrishna Rajanna)
 */
public interface AssistedInjectTargetVisitor<T, V> extends BindingTargetVisitor<T, V> {
  
  /**
   * Visits an {@link AssistedInjectBinding} created through {@link FactoryModuleBuilder}.
   */
  V visit(AssistedInjectBinding<? extends T> assistedInjectBinding);
}
