/**
 * Copyright (C) 2011 Google Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.inject.internal;

/**
 * Something that needs some delayed initialization, typically
 * a binding or internal factory that needs to be created & put
 * into the bindings map & then initialized later.
 * 
 * @author sameb@google.com (Sam Berlin)
 */
interface DelayedInitialize {
  
  /** Initializes this binding, throwing any errors if necessary. */
  void initialize(InjectorImpl injector, Errors errors) throws ErrorsException;

}
