/**
 * Copyright (C) 2008 Google Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.inject.grapher.demo;

import com.google.inject.Inject;
import com.google.inject.Provider;

class DeLorian {
  // We @Inject a Provider to demonstrate that the graph doesn't differentiate
  // between a Provider<T> and just @Injecting T. 
  @Inject @Driver Provider<Person> driver;
  @Inject FluxCapacitor fluxCapacitor;
  @Inject PrivateTestModule.Exposed exposed;
  
  @Inject
  public void setEnergySource(EnergySource energySource) {}
}
